/*	mtr.cpp -- MoToR controls
*
*	Written by Cathy Saxton
*	cathy@idleloop.com
*/

#include <avr\io.h>
#include <avr\interrupt.h>
#include <avr\signal.h>
#include "abc.h"
#include "debug.h"
#include "mtr.h"
#ifndef OPTICAL_DIST
#include "timer.h"
#endif

#define fRevL		0x01
#define fRevR		0x02
#define grfRevNone	0x00
#define grfRevAll	0x03

char MTR::m_fInit = fFalse;
uchar MTR::m_grfRev = grfRevNone;

#ifdef OPTICAL_DIST
short MTR::m_cdu = 0;
#else // OPTICAL_DIST
ushort MTR::m_cdu = 0;
#endif // OPTICAL_DIST

/* NOTE: ENUM/LIST #defines are in iop.h */
const IOP MTR::m_mpmtrpiop[mtrpMax] =
{
	MTRP_LIST
};

MTR::MTR()
{
	if (!m_fInit)
		Init();
}

/* set direction for outputs; this uses DDRx sfr's [special function registers]
   and sets bits to 1 to indicate output */
/* motor (PWM) connected to pin from timer/counter; we just need to
   send signals on it */
void MTR::Init() const
{
	const IOP *piop, *piopMax;

	m_fInit = fTrue;

	for (piopMax = (piop = m_mpmtrpiop) + mtrpOutMax; piop < piopMax; ++piop)
		piop->SetAsOutput();
	/* set counters as input */
	for (piopMax = m_mpmtrpiop + mtrpMax; piop < piopMax; ++piop)
		piop->SetAsInput();

#ifdef CLK_1MHZ
	/* 16 kHz / 256 (clock pulses) = 64 Hz for PWM pulse [frame ~15.6 ms] */
//	TCCR1B = 0x0B;	// 000 01 011 (last 3: clock [1MHz/64 = 16 kHz])
	/* start with motors disconnected from timer/counter output */
//	TCCR1A = 0x01;	// 00 00 00 01 (last 2: 01 = 8-bit, 10 = 9-bit, 11 = 10-bit)
//	TCCR1A = 0xA1;	// 10 10 00 01 (last 2: 01 = 8-bit, 10 = 9-bit, 11 = 10-bit)

	/* 128 kHz / 256 (clock pulses) = 512 Hz for PWM pulse [frame  ~2 ms] */
	TCCR1B = 0x0A;	// 000 01 010 (last 3: clock [1MHz/8 = 128 kHz])
#else
	/* 128 kHz / 256 (clock pulses) = 512 Hz for PWM pulse [frame  ~2 ms] */
	TCCR1B = 0x0B;	// 000 01 011 (last 3: clock [8MHz/64 = 128 kHz])
#endif
	/* start with motors disconnected from timer/counter output */
	TCCR1A = 0x01;	// 00 00 00 01 (last 2: 01 = 8-bit, 10 = 9-bit, 11 = 10-bit)

	OCR1A = OCR1B = 0x80;

	/* optical mouse input... */
	m_cdu = 0;
	/* hooked up on R motor pins: INT0, PC5 */
	/* set up rotation counters
	   trigger going low: 10
	   trigger going high: 11
	   (01 = any change)
	   INT1 (L), bits 3-2 in MCUCR
	   INT0 (R), bits 1-0 in MCUCR
	*/
#ifdef OPTICAL_DIST
	MCUCR &= ~0x0F;
	MCUCR |= 0x03;	// 11 = trigger going high; R motor = bits 1-0
	GICR |= 0x40;	// enable external interrupts for INT0

	/* REVIEW: should this be elsewhere? (e.g. main()) [also in timer] */
	/* set I-bit in SREG (global enable for interrupts) */
	sei();	// clear with cli();
#endif
}

void MTR::Fwd(uchar spd) const
{
	Reset(spd, grfRevNone);
}

void MTR::Fwd(uchar spdLeft, uchar spdRight) const
{
	Reset(spdLeft, spdRight, grfRevNone);
}

void MTR::Rev(uchar spd) const
{
	Reset(spd, grfRevAll);
}

void MTR::Left(uchar spd) const
{
	Reset(spd, fRevL);
}

void MTR::Right(uchar spd) const
{
	Reset(spd, fRevR);
}

void MTR::Off() const
{
	/* disconnect motors from timer/counter output
	   (00 for control for each motor) */
	TCCR1A &= ~0xF0;

	/* make sure off */
	m_mpmtrpiop[mtrpDirLeft].SignalLow();
	m_mpmtrpiop[mtrpDirRight].SignalLow();
}

/* turn on using current settings for direction
   (speed already set in OCR1A/B) */
/* note that this assumes that Off has been done before this;
   namely that we've done SignalLow for mtrpDir */
void MTR::On() const
{
	uchar grf;

	/* 10 is forward, 11 is reverse */
	/* TCCR1A high nibble is RR LL */

	grf = 0xA0;	// 10 10

	if (m_grfRev & fRevL)
	{
		grf |= 0x10;	// xx x1
		m_mpmtrpiop[mtrpDirLeft].SignalHigh();
	}

	if (m_grfRev & fRevR)
	{
		grf |= 0x40;	// x1 xx
		m_mpmtrpiop[mtrpDirRight].SignalHigh();
	}

//	TCCR1A &= ~0xF0;
//	TCCR1A |= grf;
	TCCR1A = grf | (TCCR1A & 0xf);
}

void MTR::SetSpeed(uchar spd) const
{
	OCR1A = OCR1B = spd;
}

void MTR::Reset(uchar spd, uchar grfRev) const
{
	Reset(spd, spd, grfRev);
}

void MTR::Reset(uchar spdLeft, uchar spdRight, uchar grfRev) const
{
	if ((TCCR1A & 0xF0) != 0)
	{
		/* already running */
		if (grfRev == m_grfRev)
		{
			/* same direction; just change speed */
			OCR1A = spdRight;
			OCR1B = spdLeft;
			return;
		}
		/* turn off for direction change */
		Off();
	}
	OCR1A = spdRight;
	OCR1B = spdLeft;
	m_grfRev = grfRev;
	On();
}

#ifdef NO
/* REVIEW: debug... */
#include "do.h"
void MTR::TestMotors() const
{
	TIMER timer;

	Fwd(0xF0);	// fwd 1
	timer.Wait(1);
	Off();		// off 1
	timer.Wait(1);
	On();		// fwd 1
	timer.Wait(1);
	Left(0x80);	// left 1
	timer.Wait(1);
	Off();		// off 1
	timer.Wait(1);
	On();		// left 1
	timer.Wait(1);
	Off();


	DO do_;

	Fwd(0xFF);
	do_.Blink(dopLedGreen, 3);
	Left(0xC0);
	do_.Blink(dopLedGreen, 2);
	Right(0xC0);
	do_.Blink(dopLedGreen, 1);
	Fwd(0xF0);
	do_.Blink(dopLedGreen, 2);
	Off();
}
#endif

#ifdef OPTICAL_DIST
INTERRUPT(SIG_INTERRUPT0)	// right
{
	MTR mtr;

	mtr.Count();
}

void MTR::Count() const
{
	if (m_mpmtrpiop[mtrpQuadDir].FLow())
		--m_cdu;
	else
		++m_cdu;
}
#endif // OPTICAL_DIST

void MTR::ClearDistance() const
{
#ifdef OPTICAL_DIST
	m_cdu = 0;
#else // OPTICAL_DIST
	TIMER timer;

	m_cdu = timer.CtickCur();
#endif // OPTICAL_DIST
}

/* returns distance traveled since last clear */
short MTR::Distance() const
{
#ifdef OPTICAL_DIST
	return m_cdu;
#else // OPTICAL_DIST
	ushort ctick;
	TIMER timer;

	ctick = timer.CtickCur();
	if (ctick > m_cdu)
		return ctick - m_cdu;
	return 0xffff - m_cdu + 1 + ctick;
#endif // OPTICAL_DIST
}
